// ;-*-C++-*- *  Time-stamp: "2011-03-24 01:42:52 hmmr"
/*
 *       File name:  core/iface.h
 *         Project:  Aghermann
 *          Author:  Andrei Zavada (johnhommer@gmail.com)
 * Initial version:  2008-07-01
 *
 *         Purpose:  C++ core structures and functions made accessible for ui
 *
 *         License:  GPL
 */


#ifndef _AGH_IFACE_H
#define _AGH_IFACE_H

#include <time.h>
#include <unistd.h>
#include <gsl/gsl_siman.h>
#include "../common.h"

#if HAVE_CONFIG_H
#  include "config.h"
#endif



#ifdef __cplusplus
extern "C" {
#endif



// api: expdesign

typedef void (*TProgressIndicatorFun) (const char*, size_t, size_t);

typedef void* TRecRef;
typedef void* TModelRef;

struct SEpisode {  // is synonymous with C++ class in CSubject, but nevermind
	const char
		*name;
	size_t	length;  // in seconds
	time_t	start, start_rel,
		end, end_rel;
	size_t	n_recordings;
	TRecRef	*recordings;
};

struct SModelRun {
	float	from, upto;
	TModelRef
		modref;
};

struct SModelRunSet {
	const char
		*channel;
	size_t	n_modruns;
	struct SModelRun*
		modruns;
};

struct SSession {
	const char
		*name;
	size_t n_episodes;
	struct SEpisode *episodes;
	size_t n_modrun_sets;
	struct SModelRunSet *modrun_sets;
};

struct SSubject {
	const char
		*name;
	TGender gender;
	int age;
	const char *comment;
	size_t n_sessions;
	struct SSession *sessions;
};

struct SGroup {
	const char
		*name;
	size_t	n_subjects;
	struct SSubject *subjects;
};

struct SExpDesign {
	const char
		*session_dir;
	size_t	n_groups;
	struct SGroup *groups;
};

int		agh_expdesign_init( const char* dir, TProgressIndicatorFun fun);
void		agh_expdesign_shutdown();
int		agh_expdesign_status();
const char*	agh_expdesign_messages();
const char*	agh_expdesign_last_message();
void		agh_expdesign_scan_tree( TProgressIndicatorFun);

void		agh_expdesign_snapshot( struct SExpDesign*);
void		agh_SExpDesign_destruct( struct SExpDesign*);



// enumerate used names
size_t	agh_enumerate_groups( char***);
size_t	agh_enumerate_subjects( char***);
size_t	agh_enumerate_sessions( char***);
size_t	agh_enumerate_episodes( char***);
size_t	agh_enumerate_all_channels( char***);
size_t	agh_enumerate_eeg_channels( char***);
void	agh_free_enumerated_array( char**);


// edf sources
struct SSignal_lite {
	const char
		*Channel,
		*SignalType;
	size_t	SamplesPerRecord;
};
struct SEDFFile {

	int	status;
	const char*
		filename;

      // static fields (raw)
	const char
		*PatientID,
		*Session,
		*Episode;
	size_t	NDataRecords,
		DataRecordSize,
		NSignals;
	struct tm
		timestamp_struct;
	time_t	start_time,
		end_time;

	struct SSignal_lite*
		signals;
};

void	agh_SEDFFile_destruct( struct SEDFFile*);

typedef void* TEDFRef;

// get info on a yet unattached edf file; will allocate the SEDFFile object returned
// use agh_SEDFFile_free to free the object
struct SEDFFile*	agh_edf_get_info_from_file( const char* fname, char** recp) __attribute__ ((malloc));
// get info on an edf that has been registered; will reuse a static object:
// DON'T call agh_SEDFFile_free on it!
const struct SEDFFile*	agh_edf_get_info_from_sourceref( TEDFRef, char**);

const struct SEDFFile*	agh_edf_find_by_jdeh( const char* j, const char* d, const char* e, const char* h,
					      TEDFRef*);
const struct SEDFFile*	agh_edf_find_first( TEDFRef*);
const struct SEDFFile*	agh_edf_find_next( TEDFRef*);
//int	agh_edf_write_header( TEDFRef);
int	agh_edf_export_signal( TEDFRef _F,
			       const char *h,
			       const char *fname,
			       int do_original);
size_t	agh_edf_get_scores( TEDFRef, char**, size_t *pagesize_p);
int 	agh_edf_put_scores( TEDFRef, char*);
float	agh_edf_get_percent_scored( TEDFRef);
float	agh_edf_get_scored_stages_breakdown( TEDFRef _F, float *nrem_p, float *rem_p, float *wake_p);
int	agh_edf_import_scores( TEDFRef, const char *fname);
int	agh_edf_import_scores_custom( TEDFRef, const char *fname, const char *custom_score_codes[8]);
int	agh_edf_export_scores( TEDFRef, const char *fname);

int	agh_edf_run_scoring_assistant( TEDFRef);


float		agh_edf_get_highpass_cutoff( TEDFRef, const char *h);
unsigned	agh_edf_get_highpass_order( TEDFRef, const char *h);
float		agh_edf_get_lowpass_cutoff( TEDFRef, const char *h);
unsigned	agh_edf_get_lowpass_order( TEDFRef, const char *h);
void	agh_edf_set_highpass_cutoff( TEDFRef, const char *h, float);
void	agh_edf_set_highpass_order( TEDFRef, const char *h, unsigned);
void	agh_edf_set_lowpass_cutoff( TEDFRef, const char *h, float);
void	agh_edf_set_lowpass_order( TEDFRef, const char *h, unsigned);

struct SUnfazer {
	const char *channel;
	double factor;
};

size_t	agh_edf_get_unfazers( TEDFRef,
			      const char *affected_channel,
			      struct SUnfazer **recp);
int	agh_edf_add_or_mod_unfazer( TEDFRef,
				    const char *affected_channel,
				    const char *offending_channel,
				    double factor);
double	agh_edf_get_unfazer_factor( TEDFRef, const char*, const char*);
void	agh_edf_remove_unfazer( TEDFRef, const char*, const char*);

size_t	agh_edf_get_artifacts( TEDFRef, const char *channel, size_t**);
void	agh_edf_mark_artifact( TEDFRef, const char*, size_t, size_t);
void	agh_edf_clear_artifact( TEDFRef, const char*, size_t, size_t);


void	agh_explain_edf_status( int status, char **out_p);



// group

const char*	agh_group_find_first();
const char*	agh_group_find_next();


// subject

size_t	agh_subject_get_n_of();
size_t	agh_subject_get_n_of_in_group( const char*);
size_t	agh_subject_get_path( const char *j, char **outp);
// if you pass a recp argument for the agh_subject_find_* functions
// below, make sure you call agh_SSubject_destruct on it when unneeded
// or before the next call to agh_subject_find_*
const struct SSubject*	agh_subject_find_by_name( const char* j_name, struct SSubject*);
const struct SSubject*	agh_subject_find_first_in_group( const char* which_group, struct SSubject*);
const struct SSubject*	agh_subject_find_next_in_group( struct SSubject*);

void	agh_SSubject_destruct( struct SSubject*);


struct SEpisodeTimes {
	int start_hour, start_min, start_sec;
	int end_hour, end_min, end_sec;
};

	// return duration, h
float	agh_group_avg_episode_times( const char *group, const char *session, const char *episode,
				     struct SEpisodeTimes *recp);


// measurement

TRecRef		agh_msmt_find_by_jdeh( const char *j,
				       const char *d,
				       const char *e,
				       const char *h);
TEDFRef		agh_msmt_get_source( TRecRef);
const char*	agh_msmt_get_signal_type( TRecRef);
const char*	agh_msmt_get_signal_name( TRecRef);
size_t		agh_msmt_get_pagesize( TRecRef);
float		agh_msmt_get_binsize( TRecRef);
size_t		agh_msmt_get_n_bins( TRecRef);

size_t		agh_msmt_get_signal_original_as_double( TRecRef,
							double **out_p,
							size_t *samplerate,
							float *signal_scale);
size_t		agh_msmt_get_signal_original_as_float( TRecRef,
						       float **out_p,
						       size_t *samplerate,
						       float *signal_scale);
size_t		agh_msmt_get_signal_filtered_as_double( TRecRef,
							double **out_p,
							size_t *samplerate,
							float *signal_scale);
size_t		agh_msmt_get_signal_filtered_as_float( TRecRef,
						       float **out_p,
						       size_t *samplerate,
						       float *signal_scale);

// msmt continued
size_t		agh_msmt_get_power_spectrum_as_double( TRecRef, size_t p, double**, double *max_p);
size_t		agh_msmt_get_power_spectrum_as_float( TRecRef, size_t p, float**, float*);

size_t		agh_msmt_get_power_course_as_double_direct( TRecRef, double*);
size_t		agh_msmt_get_power_course_as_float_direct( TRecRef, float*);
size_t		agh_msmt_get_power_course_as_double( TRecRef, double**);
size_t		agh_msmt_get_power_course_as_float( TRecRef, float**);
size_t		agh_msmt_get_power_course_in_range_as_double( TRecRef, float, float,
							      double**);
size_t		agh_msmt_get_power_course_in_range_as_float( TRecRef, float, float,
							     float**);
void		agh_msmt_get_power_course_in_range_as_double_direct( TRecRef, float, float,
								     double*);
void		agh_msmt_get_power_course_in_range_as_float_direct( TRecRef, float, float,
								    float*);
char*		agh_msmt_fname_base( TRecRef)  __attribute__ ((malloc));
int		agh_msmt_export_power( TRecRef, const char *fname);
int		agh_msmt_export_power_in_range( TRecRef, float, float, const char *fname);


int		agh_episode_assisted_score_by_jde( const char *j_name, const char *d_name, const char *e_name);


size_t		agh_fft_get_pagesize();
void		agh_fft_set_pagesize( size_t);
float		agh_fft_get_binsize();
void		agh_fft_set_binsize( float);
TFFTWinType	agh_fft_get_window_type();
void		agh_fft_set_window_type( TFFTWinType);

TFFTWinType	agh_af_get_window_type();
void		agh_af_set_window_type( TFFTWinType);



TModelRef	agh_modelrun_find_by_jdhq( const char *j_name, const char *d_name, const char *h_name,
					   float from, float upto);

const char*	agh_modelrun_get_subject( TModelRef);
const char*	agh_modelrun_get_session( TModelRef);
const char*	agh_modelrun_get_channel( TModelRef);
float		agh_modelrun_get_freqrange( TModelRef, float*, float*);
size_t		agh_modelrun_get_n_episodes( TModelRef);
size_t		agh_modelrun_get_nth_episode_start_p( TModelRef, size_t);
size_t		agh_modelrun_get_nth_episode_end_p( TModelRef, size_t);
size_t		agh_modelrun_get_pagesize( TModelRef);


//const double*	agh_modelrun_get_timeline( TModelRef, size_t *length_p);
size_t		agh_modelrun_get_all_courses_as_double( TModelRef,
							double **SWA_out, double **S_out, double **SWAsim_out,
							char **scores_out);
size_t		agh_modelrun_get_mutable_courses_as_double( TModelRef,
							    double **S_out, double **SWAsim_out);
void		agh_modelrun_get_all_courses_as_double_direct( TModelRef Ri,
							       double *SWA_out, double *S_out, double *SWAsim_out,
							       char *scores_out);
void		agh_modelrun_get_mutable_courses_as_double_direct( TModelRef Ri,
								   double *S_out, double *SWAsim_out);
int		agh_modelrun_setup( const char *j, const char *d, const char *h,
				    float from, float upto,
				    TModelRef*,
				    const char **error_p);
struct SConsumerSCourseSetupInfo {
	size_t	_sim_start,
		_sim_end,
		_baseline_end,
		_pages_with_SWA,
		_pages_in_bed;
	double	_SWA_L,
		_SWA_0,	_SWA_100;
};

const struct SConsumerSCourseSetupInfo *
		agh_modelrun_get_scourse_setup_info( TModelRef, struct SConsumerSCourseSetupInfo*);

int		agh_modelrun_run( TModelRef, void (*)(void*));
double		agh_modelrun_snapshot( TModelRef);  // do a single cycle to recreate variable courses

void		agh_modelrun_remove_untried();

int		agh_modelrun_tsv_export_one( TModelRef, const char *fname);
int		agh_modelrun_tsv_export_all( const char *fname);


struct SConsumerTunableSet {
	size_t	n_tunables;
	double	tunables[_agh_n_tunables_];
};

void		agh_modelrun_get_tunables( TModelRef Ri, struct SConsumerTunableSet*);
void		agh_modelrun_put_tunables( TModelRef Ri, const struct SConsumerTunableSet*);

struct SConsumerCtlParams {
	gsl_siman_params_t
		siman_params;
	int	DBAmendment1:1,
		DBAmendment2:1,
		AZAmendment:1,
		ScoreMVTAsWake:1,
		ScoreUnscoredAsWake:1;
	float	req_percent_scored;
	size_t	swa_laden_pages_before_SWA_0;
};

void		agh_modelrun_get_ctlparams( TModelRef Ri, struct SConsumerCtlParams*);
void		agh_modelrun_put_ctlparams( TModelRef Ri, const struct SConsumerCtlParams*);

int		agh_modelrun_get_status( TModelRef Ri);

const struct STunableDescription*
		agh_tunable_get_description( size_t);


struct SConsumerTunableSetFull {
	size_t	n_tunables;  // # actually used
	double	tunables[_agh_n_tunables_],
		upper_bounds[_agh_n_tunables_],
		lower_bounds[_agh_n_tunables_],
		steps[_agh_n_tunables_];
	int	states[_agh_n_tunables_];
};

void		agh_tunables0_get( struct SConsumerTunableSetFull*, size_t);
void		agh_tunables0_put( const struct SConsumerTunableSetFull*, size_t);
void		agh_tunables0_stock_defaults();

void		agh_ctlparams0_get( struct SConsumerCtlParams*);
void		agh_ctlparams0_put( const struct SConsumerCtlParams*);

float		agh_req_percent_scored_get();
void		agh_req_percent_scored_set( float);



int		agh_channel_follows_1020( const char* h);
const char*	agh_signal_type_following_Kemp( const char* h);
int		agh_signal_type_is_fftable( const char*);



#ifdef __cplusplus
}
#endif


#endif

// EOF
